@@ -1,15 +1,15 @@
 """Cyberbrain public API and tracer setup."""
-import argparse
 import dis
+
+import argparse
 import functools
 import inspect
 import os
+import requests
 import sys
 from types import MethodType, FunctionType, FrameType
 from typing import Optional, Union
 
-import grpc
-
 from . import logger, utils, rpc_client
 from .frame import Frame
 from .frame_tree import FrameTree
@@ -58,11 +58,18 @@ class Tracer:
         self.is_generator_function = False
         self.decorated_function_code_id = None
         self.frame_logger: Optional[logger.FrameLogger] = None
+        self.tracer_state = TracerFSM.INITIAL
         if debug_mode is not None:
             self.debug_mode = debug_mode
 
-        self.tracer_state = TracerFSM.INITIAL
-        self.rpc_client = rpc_client.RpcClient()
+        if utils.run_in_test():
+            import portpicker
+
+            self.rpc_client = rpc_client.RpcClient(
+                rpc_server_port=portpicker.pick_unused_port()
+            )
+        else:
+            self.rpc_client = rpc_client.RpcClient()
 
     def _initialize_frame_and_logger(
         self, raw_frame: FrameType, initial_instr_pointer: int
@@ -141,9 +148,8 @@ class Tracer:
             assert len(self.frame_logger.frame.value_stack.stack) == 0
 
         try:
-            # print("call send")
             self.rpc_client.send_frame(self.frame)
-        except grpc._channel._InactiveRpcError:
+        except requests.exceptions.ConnectionError:
             print("Can't connect to RPC server")
 
     def __call__(self, disabled: Union[Union[FunctionType, MethodType], bool] = False):
@@ -233,9 +239,12 @@ class Tracer:
     def local_tracer(self, raw_frame, event, arg):
         if utils.should_exclude(raw_frame):
             return
+        if event == "exception":
+            # print("Exception 🐍", raw_frame.f_lasti, arg)
+            self.frame_logger.handle_exception(arg)
         if event == "opcode":
             # print(raw_frame, event, arg, raw_frame.f_lasti)
-            self.frame_logger.update(raw_frame)
+            self.frame_logger.handle_instructions(raw_frame)
         if event == "return":
             # print(raw_frame, event, arg, raw_frame.f_lasti)
             self.frame.log_return_event(raw_frame, value=arg)
