@@ -2,6 +2,7 @@
 
 from __future__ import annotations
 
+import argparse
 import dis
 import inspect
 import os
@@ -15,13 +16,15 @@ from pprint import pformat
 from types import FrameType
 from typing import Any, Optional
 
+import cheap_repr
 import jsonpickle
 import more_itertools
-from cheap_repr import cheap_repr
 from pygments import highlight
 from pygments.formatters import Terminal256Formatter
 from pygments.lexers import PythonLexer
 
+from . import tracer
+
 # These pickle handlers rely on Numpy and Pandas, so it only makes sense to register
 # them if Numpy or Pandas are installed.
 try:
@@ -43,15 +46,22 @@ except ImportError:
 _INSTALLATION_PATHS = list(sysconfig.get_paths().values())
 _PYTHON_EXECUTABLE_PATH = sys.executable
 
+jsonpickle.set_preferred_backend("ujson")
+
+
+# To not let it show warnings
+@cheap_repr.register_repr(argparse.Namespace)
+def repr_for_namespace(_, __):
+    return "argparse.Namespace"
+
 
 def should_ignore_event(
     *, target: str, value: Any, frame: Optional[FrameType] = None
 ) -> bool:
     """Determines whether we should ignore this event."""
-    from .tracer import Tracer
 
     # Excludes events from tracer.
-    if isinstance(value, Tracer):
+    if isinstance(value, tracer.Tracer):
         return True
 
     # Excludes events from modules.
@@ -92,7 +102,13 @@ def get_jump_target_or_none(instr: dis.Instruction) -> Optional[int]:
 
 def to_json(python_object: Any):
     # TODO: Once we implemented better deserialization in Js, use unpicklable=True.
-    json = jsonpickle.encode(python_object, unpicklable=False)
+    try:
+        json = jsonpickle.encode(python_object, unpicklable=False)
+    except:
+        # There are always things we just cannot encode, like a ML model.
+        # In this case, use its repr.
+        return '{"repr": "%s"}' % get_repr(python_object)
+
     if json == "null" and python_object is not None:
         return '{"repr": "%s"}' % get_repr(python_object)
     else:
@@ -213,7 +229,7 @@ def get_repr(obj: Any) -> str:
     if type(obj) == str:
         return f'"{obj}"'
 
-    repr_text = cheap_repr(obj)
+    repr_text = cheap_repr.cheap_repr(obj)
     match = re.search("at 0x", repr_text)
     if not match:
         return repr_text
