@@ -1,3 +1,5 @@
+# encoding: utf-8
+
 """A self maintained value stack."""
 
 from __future__ import annotations
@@ -18,7 +20,14 @@ except ImportError:
     from typing_extensions import Literal
 
 from . import utils
-from .basis import Symbol, Binding, Mutation, Deletion, JumpBackToLoopStart
+from .basis import (
+    Symbol,
+    Binding,
+    Mutation,
+    Deletion,
+    JumpBackToLoopStart,
+    ExceptionInfo,
+)
 from .block_stack import BlockStack, BlockType, Block
 
 if TYPE_CHECKING:
@@ -43,13 +52,6 @@ class _NullClass:
 NULL = _NullClass()
 
 
-@dataclasses.dataclass
-class ExceptionInfo:
-    type: type
-    value: Exception
-    traceback: any
-
-
 def emit_event(f):
     """Decorator used to denote that a handler emits at least one event.
 
@@ -74,10 +76,21 @@ class EventInfo:
     jump_target: int = None
 
 
-class GeneralValueStack:
+class Why(enum.Enum):
+    UNINITIALIZED = 0
+    NOT = 1  # No error
+    EXCEPTION = 2  # Exception occurred
+    RETURN = 3  # 'return' statement
+    BREAK = 4  # 'break' statement
+    CONTINUE = 5  # 'continue' statement
+    YIELD = 6  # 'yield' operator
+    SILENCED = 8  # Exception silenced by 'with'
+
+
+class BaseValueStack:
     """Class that simulates the a frame's value stack.
 
-    This class handles instructions that don't require special processing.
+    This class contains instr handlers that are the same across different versions.
     """
 
     def __init__(self):
@@ -85,6 +98,7 @@ class GeneralValueStack:
         self.block_stack = BlockStack()
         self.last_exception: Optional[ExceptionInfo] = None
         self.return_value = _placeholder
+        self.handler_signature_cache: dict[str, set[str]] = {}  # keyed by opname.
         self.snapshot = None
 
     def update_snapshot(self, mutated_identifier: str, new_snapshot: Snapshot):
@@ -104,7 +118,12 @@ class GeneralValueStack:
                     symbol.snapshot = new_snapshot
 
     def emit_event_and_update_stack(
-        self, instr: Instruction, frame: FrameType, jumped: bool, snapshot: Snapshot
+        self,
+        instr: Instruction,
+        frame: FrameType,
+        jumped: bool,
+        exc_info: Optional[ExceptionInfo],
+        snapshot: Snapshot,
     ) -> Optional[EventInfo]:
         """Given a instruction, emits EventInfo if any, and updates the stack.
 
@@ -112,31 +131,41 @@ class GeneralValueStack:
             instr: current instruction.
             jumped: whether jump just happened.
             frame: current frame.
+            exc_info: implicitly raised exception if any, or None.
             snapshot: frame state snapshot.
         """
         self.snapshot = snapshot
+        opname = instr.opname
 
-        if instr.opname.startswith("BINARY") or instr.opname.startswith("INPLACE"):
+        if opname.startswith("BINARY") or opname.startswith("INPLACE"):
             # Binary operations are all the same.
             handler = self._BINARY_operation_handler
         else:
             try:
-                handler = getattr(self, f"_{instr.opname}_handler")
+                handler = getattr(self, f"_{opname}_handler")
             except AttributeError:
-                raise AttributeError(
-                    f"Please add\ndef _{instr.opname}_handler(self, instr):"
-                )
+                raise AttributeError(f"Please add\ndef _{opname}_handler(self, instr):")
 
         # Pass arguments on demand.
-        parameters = inspect.signature(handler).parameters
-        args = []
-        if "instr" in parameters:
-            args.append(instr)
-        if "jumped" in parameters:
-            args.append(jumped)
-        if "frame" in parameters:
-            args.append(frame)
-        return handler(*args)
+        try:
+            parameters = self.handler_signature_cache[opname]
+        except KeyError:
+            parameters = set(inspect.signature(handler).parameters)
+            self.handler_signature_cache[opname] = parameters
+
+        # noinspection PyArgumentList
+        return handler(
+            *[
+                arg
+                for param_name, arg in {
+                    "instr": instr,
+                    "jumped": jumped,
+                    "frame": frame,
+                    "exc_info": exc_info,
+                }.items()
+                if param_name in parameters
+            ]
+        )
 
     @property
     def stack_level(self):
@@ -234,6 +263,18 @@ class GeneralValueStack:
         while self.stack_level > b.b_level:
             self._pop()
 
+    def _instruction_successfully_executed(
+        self, exc_info: Optional[ExceptionInfo], opname: str
+    ) -> bool:
+        """Returns true if there's no exception, otherwise false."""
+        if exc_info:
+            sys.stdout.buffer.write(
+                f"⚠️  Exception happened in {opname}\n".encode("utf-8")
+            )
+            self._store_exception(exc_info)
+            return False
+        return True
+
     def _POP_TOP_handler(self):
         self._pop()
 
@@ -269,21 +310,29 @@ class GeneralValueStack:
     def _UNARY_INVERT_handler(self, instr):
         pass
 
-    def _BINARY_operation_handler(self):
-        self._pop_n_push_one(2)
+    def _BINARY_operation_handler(self, exc_info):
+        tos, tos1 = self._pop(2)
+        if self._instruction_successfully_executed(exc_info, "BINARY op"):
+            self._push(utils.flatten(tos, tos1))
 
     @emit_event
-    def _STORE_SUBSCR_handler(self):
+    def _STORE_SUBSCR_handler(self, exc_info):
         tos, tos1, tos2 = self._pop(3)
-        assert len(tos1) == 1
-        return EventInfo(type=Mutation, target=tos1[0], sources=set(tos + tos1 + tos2))
+        if self._instruction_successfully_executed(exc_info, "STORE_SUBSCR"):
+            # We use to `assert len(tos1) == 1`, but in certain cases, like
+            # os.environ["foo"] = "2", tos1 is [].
+            if tos1:
+                return EventInfo(
+                    type=Mutation, target=tos1[0], sources=set(tos + tos1 + tos2)
+                )
 
     # noinspection DuplicatedCode
     @emit_event
-    def _DELETE_SUBSCR_handler(self):
+    def _DELETE_SUBSCR_handler(self, exc_info):
         tos, tos1 = self._pop(2)
         assert len(tos1) == 1
-        return EventInfo(type=Mutation, target=tos1[0], sources=set(tos + tos1))
+        if self._instruction_successfully_executed(exc_info, "DELETE_SUBSCR"):
+            return EventInfo(type=Mutation, target=tos1[0], sources=set(tos + tos1))
 
     def _YIELD_VALUE_handler(self):
         """
@@ -319,38 +368,47 @@ class GeneralValueStack:
         return binding
 
     @emit_event
-    def _DELETE_NAME_handler(self, instr):
-        return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
+    def _DELETE_NAME_handler(self, instr, exc_info):
+        if self._instruction_successfully_executed(exc_info, "DELETE_NAME"):
+            return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
 
-    def _UNPACK_SEQUENCE_handler(self, instr):
-        self._pop_one_push_n(instr.arg)
+    def _UNPACK_SEQUENCE_handler(self, instr, exc_info):
+        seq = self._pop()
+        if self._instruction_successfully_executed(exc_info, "UNPACK_SEQUENCE"):
+            for _ in range(instr.arg):
+                self._push(seq)
 
-    def _UNPACK_EX_handler(self, instr):
+    def _UNPACK_EX_handler(self, instr, exc_info):
         assert instr.arg <= 65535  # At most one extended arg.
         higher_byte, lower_byte = instr.arg >> 8, instr.arg & 0x00FF
         number_of_receivers = lower_byte + 1 + higher_byte
-        self._pop_one_push_n(number_of_receivers)
+        seq = self._pop()
+        if self._instruction_successfully_executed(exc_info, "UNPACK_EX"):
+            for _ in range(number_of_receivers):
+                self._push(seq)
 
-    # noinspection DuplicatedCode
     @emit_event
-    def _STORE_ATTR_handler(self):
+    def _STORE_ATTR_handler(self, exc_info):
         tos, tos1 = self._pop(2)
         assert len(tos) == 1
-        return EventInfo(type=Mutation, target=tos[0], sources=set(tos + tos1))
+        if self._instruction_successfully_executed(exc_info, "STORE_ATTR"):
+            return EventInfo(type=Mutation, target=tos[0], sources=set(tos + tos1))
 
     @emit_event
-    def _DELETE_ATTR_handler(self):
+    def _DELETE_ATTR_handler(self, exc_info):
         tos = self._pop()
         assert len(tos) == 1
-        return EventInfo(type=Mutation, target=tos[0], sources=set(tos))
+        if self._instruction_successfully_executed(exc_info, "DELETE_ATTR"):
+            return EventInfo(type=Mutation, target=tos[0], sources=set(tos))
 
     @emit_event
     def _STORE_GLOBAL_handler(self, instr):
         return self._STORE_NAME_handler(instr)
 
     @emit_event
-    def _DELETE_GLOBAL_handler(self, instr):
-        return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
+    def _DELETE_GLOBAL_handler(self, instr, exc_info):
+        if self._instruction_successfully_executed(exc_info, "DELETE_GLOBAL"):
+            return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
 
     def _BUILD_TUPLE_handler(self, instr):
         self._pop_n_push_one(instr.arg)
@@ -358,11 +416,15 @@ class GeneralValueStack:
     def _BUILD_LIST_handler(self, instr):
         self._BUILD_TUPLE_handler(instr)
 
-    def _BUILD_SET_handler(self, instr):
-        self._BUILD_TUPLE_handler(instr)
+    def _BUILD_SET_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(exc_info, "BUILD_SET"):
+            self._push(utils.flatten(items))
 
-    def _BUILD_MAP_handler(self, instr):
-        self._pop_n_push_one(instr.arg * 2)
+    def _BUILD_MAP_handler(self, instr, exc_info):
+        items = self._pop(instr.arg * 2)
+        if self._instruction_successfully_executed(exc_info, "BUILD_MAP"):
+            self._push(utils.flatten(items))
 
     def _BUILD_CONST_KEY_MAP_handler(self, instr):
         self._pop_n_push_one(instr.arg + 1)
@@ -370,25 +432,41 @@ class GeneralValueStack:
     def _BUILD_STRING_handler(self, instr):
         self._pop_n_push_one(instr.arg)
 
-    def _BUILD_TUPLE_UNPACK_handler(self, instr):
-        self._pop_n_push_one(instr.arg)
+    def _BUILD_TUPLE_UNPACK_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(exc_info, "BUILD_TUPLE_UNPACK"):
+            self._push(utils.flatten(items))
 
-    def _BUILD_TUPLE_UNPACK_WITH_CALL_handler(self, instr):
-        self._pop_n_push_one(instr.arg)
+    def _BUILD_TUPLE_UNPACK_WITH_CALL_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(
+            exc_info, "BUILD_TUPLE_UNPACK_WITH_CALL"
+        ):
+            self._push(utils.flatten(items))
 
-    def _BUILD_LIST_UNPACK_handler(self, instr):
-        self._pop_n_push_one(instr.arg)
+    def _BUILD_LIST_UNPACK_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(exc_info, "BUILD_LIST_UNPACK"):
+            self._push(utils.flatten(items))
 
-    def _BUILD_SET_UNPACK_handler(self, instr):
-        self._pop_n_push_one(instr.arg)
+    def _BUILD_SET_UNPACK_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(exc_info, "BUILD_SET_UNPACK"):
+            self._push(utils.flatten(items))
 
-    def _BUILD_MAP_UNPACK_handler(self, instr):
-        self._pop_n_push_one(instr.arg)
+    def _BUILD_MAP_UNPACK_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(exc_info, "BUILD_MAP_UNPACK"):
+            self._push(utils.flatten(items))
 
-    def _BUILD_MAP_UNPACK_WITH_CALL_handler(self, instr):
-        self._pop_n_push_one(instr.arg)
+    def _BUILD_MAP_UNPACK_WITH_CALL_handler(self, instr, exc_info):
+        items = self._pop(instr.arg)
+        if self._instruction_successfully_executed(
+            exc_info, "BUILD_MAP_UNPACK_WITH_CALL"
+        ):
+            self._push(utils.flatten(items))
 
-    def _LOAD_ATTR_handler(self, instr):
+    def _LOAD_ATTR_handler(self, exc_info):
         """Event the behavior of LOAD_ATTR.
 
         The effect of LOAD_ATTR is: Replaces TOS with getattr(TOS, co_names[namei]).
@@ -421,29 +499,34 @@ class GeneralValueStack:
         to us, so it's fine if we don't store it. But the "name" of the source object is
         vital, so we need to keep it. Thus, the handler just does nothing.
         """
+        self._instruction_successfully_executed(exc_info, "LOAD_ATTR")
 
-    def _COMPARE_OP_handler(self):
-        return self._BINARY_operation_handler()
+    def _COMPARE_OP_handler(self, exc_info):
+        return self._BINARY_operation_handler(exc_info)
 
-    def _IMPORT_NAME_handler(self):
-        self._pop_n_push_one(2)
+    def _IMPORT_NAME_handler(self, exc_info):
+        self._pop(2)
+        if self._instruction_successfully_executed(exc_info, "IMPORT_NAME"):
+            self._push(_placeholder)
 
-    def _IMPORT_FROM_handler(self):
-        self._push(_placeholder)
-
-    ############################ LOAD instructions ############################
+    def _IMPORT_FROM_handler(self, exc_info):
+        if self._instruction_successfully_executed(exc_info, "IMPORT_FROM"):
+            self._push(_placeholder)
 
     def _LOAD_CONST_handler(self):
         self._push(_placeholder)
 
-    def _LOAD_NAME_handler(self, instr, frame):
-        self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
+    def _LOAD_NAME_handler(self, instr, frame, exc_info):
+        if self._instruction_successfully_executed(exc_info, "LOAD_NAME"):
+            self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
 
-    def _LOAD_GLOBAL_handler(self, instr, frame):
-        self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
+    def _LOAD_GLOBAL_handler(self, instr, frame, exc_info):
+        if self._instruction_successfully_executed(exc_info, "LOAD_GLOBAL"):
+            self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
 
-    def _LOAD_FAST_handler(self, instr, frame):
-        self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
+    def _LOAD_FAST_handler(self, instr, frame, exc_info):
+        if self._instruction_successfully_executed(exc_info, "LOAD_FAST"):
+            self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
 
     def _fetch_value_for_load_instruction(self, name, frame):
         """Transforms the value to be loaded onto value stack based on their types.
@@ -469,29 +552,45 @@ class GeneralValueStack:
     def _STORE_FAST_handler(self, instr):
         return self._STORE_NAME_handler(instr)
 
-    def _LOAD_CLOSURE_handler(self, instr, frame):
-        self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
+    def _LOAD_CLOSURE_handler(self, instr, frame, exc_info):
+        if self._instruction_successfully_executed(exc_info, "LOAD_CLOSURE"):
+            # It is possible that the name does not exist in the frame. Example:
+            #
+            # class Bar(Foo):  # LOAD_CLOSURE, but the name `Bar` does exist in frame.
+            #     def __init__(self):
+            #         super(Bar, self).__init__()
+            #
+            # In this case, we ignore the value cause it doesn't matter.
+            try:
+                value = self._fetch_value_for_load_instruction(instr.argrepr, frame)
+            except AssertionError:
+                value = []
+            self._push(value)
 
-    def _LOAD_DEREF_handler(self, instr, frame):
-        self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
+    def _LOAD_DEREF_handler(self, instr, frame, exc_info):
+        if self._instruction_successfully_executed(exc_info, "LOAD_DEREF"):
+            self._push(self._fetch_value_for_load_instruction(instr.argrepr, frame))
 
     @emit_event
     def _STORE_DEREF_handler(self, instr):
         return self._STORE_NAME_handler(instr)
 
     @emit_event
-    def _DELETE_DEREF_handler(self, instr):
-        return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
+    def _DELETE_DEREF_handler(self, instr, exc_info):
+        if self._instruction_successfully_executed(exc_info, "DELETE_DEREF"):
+            return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
 
     @emit_event
-    def _DELETE_FAST_handler(self, instr):
-        return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
+    def _DELETE_FAST_handler(self, instr, exc_info):
+        if self._instruction_successfully_executed(exc_info, "DELETE_FAST"):
+            return EventInfo(type=Deletion, target=Symbol(instr.argrepr))
 
-    def _LOAD_METHOD_handler(self):
-        # NULL should be pushed if method lookup failed, but this would lead to an
-        # exception anyway, and should be very rare, so ignoring it.
-        # See https://docs.python.org/3/library/dis.html#opcode-LOAD_METHOD.
-        self._push(self.tos)
+    def _LOAD_METHOD_handler(self, exc_info):
+        if self._instruction_successfully_executed(exc_info, "LOAD_METHOD"):
+            # NULL should be pushed if method lookup failed, but this would lead to an
+            # exception anyway, and should be very rare, so ignoring it.
+            # See https://docs.python.org/3/library/dis.html#opcode-LOAD_METHOD.
+            self._push(self.tos)
 
     def _push_arguments_or_exception(self, callable_obj, args):
         if utils.is_exception_class(callable_obj):
@@ -503,31 +602,43 @@ class GeneralValueStack:
             # Return value is a list containing the callable and all arguments
             self._push(utils.flatten(callable_obj, args))
 
-    def _CALL_FUNCTION_handler(self, instr):
+    def _CALL_FUNCTION_handler(self, instr, exc_info):
         args = self._pop(instr.arg)
         callable_obj = self._pop()
-        self._push_arguments_or_exception(callable_obj, args)
+        if self._instruction_successfully_executed(exc_info, "CALL_FUNCTION"):
+            if utils.is_exception(args):
+                args = (args,)
+            self._push_arguments_or_exception(callable_obj, args)
 
-    def _CALL_FUNCTION_KW_handler(self, instr: Instruction):
+    def _CALL_FUNCTION_KW_handler(self, instr: Instruction, exc_info):
         args_num = instr.arg
         _ = self._pop()  # A tuple of keyword argument names.
         args = self._pop(args_num)
         callable_obj = self._pop()
-        self._push_arguments_or_exception(callable_obj, args)
+        if self._instruction_successfully_executed(exc_info, "CALL_FUNCTION_KW"):
+            if utils.is_exception(args):
+                args = (args,)
+            self._push_arguments_or_exception(callable_obj, args)
 
-    def _CALL_FUNCTION_EX_handler(self, instr):
+    def _CALL_FUNCTION_EX_handler(self, instr, exc_info):
         kwargs = self._pop() if (instr.arg & 0x01) else []
         args = self._pop()
         args.extend(kwargs)
         callable_obj = self._pop()
-        self._push_arguments_or_exception(callable_obj, args)
+        if self._instruction_successfully_executed(exc_info, "CALL_FUNCTION_EX"):
+            if utils.is_exception(args):
+                args = (args,)
+            self._push_arguments_or_exception(callable_obj, args)
 
     @emit_event
-    def _CALL_METHOD_handler(self, instr):
+    def _CALL_METHOD_handler(self, instr, exc_info):
         args = self._pop(instr.arg)
         inst_or_callable = self._pop()
         method_or_null = self._pop()  # method or NULL
-        self._push(utils.flatten(inst_or_callable, method_or_null, *args))
+        if self._instruction_successfully_executed(exc_info, "CALL_METHOD"):
+            if utils.is_exception(args):
+                args = (args,)
+            self._push(utils.flatten(inst_or_callable, method_or_null, *args))
 
         # The real callable can be omitted for various reasons.
         # See the _fetch_value_for_load method.
@@ -569,13 +680,15 @@ class GeneralValueStack:
         # Instruction.arg already contains the final value of arg, so this is a no op.
         pass
 
-    def _FORMAT_VALUE_handler(self, instr):
+    def _FORMAT_VALUE_handler(self, instr, exc_info):
         # See https://git.io/JvjTg to learn what this opcode is doing.
         elements = []
         if (instr.arg & 0x04) == 0x04:
             elements.extend(self._pop())
         elements.extend(self._pop())
-        self._push(elements)
+
+        if self._instruction_successfully_executed(exc_info, "FORMAT_VALUE"):
+            self._push(elements)
 
     def _JUMP_FORWARD_handler(self, instr, jumped):
         pass
@@ -611,35 +724,40 @@ class GeneralValueStack:
         return self._return_jump_back_event_if_exists(instr)
 
     @emit_event
-    def _GET_ITER_handler(self, instr):
-        return self._return_jump_back_event_if_exists(instr)
+    def _GET_ITER_handler(self, instr, exc_info):
+        if self._instruction_successfully_executed(exc_info, "GET_ITER"):
+            return self._return_jump_back_event_if_exists(instr)
 
     @emit_event
     def _GET_YIELD_FROM_ITER_handler(self, instr):
+        """Since the handling of generators is ad-hoc, for now we didn't handle
+        exceptions. We'll add it as part of a more well-thought-out generator
+        implementation.
+        """
         return self._return_jump_back_event_if_exists(instr)
 
     @emit_event
-    def _FOR_ITER_handler(self, instr, jumped):
-        if jumped:
-            self._pop()
+    def _FOR_ITER_handler(self, instr, jumped, exc_info: ExceptionInfo):
+        # If it's StopIteration, we assume it's OK.
+        if exc_info is None or exc_info.type is StopIteration:
+            if jumped:
+                self._pop()
+            else:
+                self._push(self.tos)
+                return self._return_jump_back_event_if_exists(instr)
         else:
-            self._push(self.tos)
-            return self._return_jump_back_event_if_exists(instr)
+            self._instruction_successfully_executed(exc_info, "FOR_ITER")
 
     def _LOAD_BUILD_CLASS_handler(self):
         self._push(_placeholder)  # builtins.__build_class__()
 
-    def _SETUP_WITH_handler(self):
-        enter_func = self.tos
-        # We ignored the operation to replace context manager on tos with __exit__,
-        # because it is a noop in our stack.
-        self._push_block(BlockType.SETUP_FINALLY)
-        self._push(enter_func)  # The return value of __enter__()
-
-    def _WITH_CLEANUP_FINISH_handler(self):
-        self._pop(2)
-        # Again, this assumes there's no unhandled exception, and ignored the handling
-        # for those exceptions.
+    def _SETUP_WITH_handler(self, exc_info):
+        if self._instruction_successfully_executed(exc_info, "SETUP_WITH"):
+            enter_func = self.tos
+            # We ignored the operation to replace context manager on tos with __exit__,
+            # because it is a noop in our stack.
+            self._push_block(BlockType.SETUP_FINALLY)
+            self._push(enter_func)  # The return value of __enter__()
 
     def _return_jump_back_event_if_exists(self, instr):
         jump_target = utils.get_jump_target_or_none(instr)
@@ -692,24 +810,37 @@ class GeneralValueStack:
         return False
 
 
-class Why(enum.Enum):
-    UNINITIALIZED = 0
-    NOT = 1  # No error
-    EXCEPTION = 2  # Exception occurred
-    RETURN = 3  # 'return' statement
-    BREAK = 4  # 'break' statement
-    CONTINUE = 5  # 'continue' statement
-    YIELD = 6  # 'yield' operator
-    SILENCED = 8  # Exception silenced by 'with'
-
-
-class Py37ValueStack(GeneralValueStack):
+class Py37ValueStack(BaseValueStack):
     """Value stack for Python 3.7."""
 
     def __init__(self):
         self.why = Why.UNINITIALIZED
         super().__init__()
 
+    def _store_exception(self, exc_info: ExceptionInfo):
+        """When an exception is raised implicitly (aka not by calling `raise`), use
+        This method to propagate it as self.last_exception.
+        """
+        self.last_exception = exc_info
+        self.why = Why.EXCEPTION
+        self._fast_block_end()
+
+    def _WITH_CLEANUP_FINISH_handler(self):
+        # For __exit__, returning a true value from this method will cause the with
+        # statement to suppress the exception and continue execution with the statement
+        # immediately following the with statement. Otherwise the exception continues
+        # propagating after this method has finished executing.
+        #
+        # res represents the return value, but there's no way CB can know its value.
+        # So we just assume res is true whenever there is an exception, because CB does
+        # not support unhandled exception, so it's safe to assume that if there's an
+        # exception raised in `with`, it is properly handled. e.g.
+        #         with pytest.raises(TypeError)
+        res = self._pop()
+        exc = self._pop()
+        if res and utils.is_exception(exc):
+            self._push(Why.SILENCED)
+
     def _RETURN_VALUE_handler(self):
         self.return_value = self._pop()
         self.why = Why.RETURN
@@ -751,6 +882,11 @@ class Py37ValueStack(GeneralValueStack):
         self._unwind_except_handler(block)
 
     def _RAISE_VARARGS_handler(self, instr):
+        # Since in FrameLogger.handle_exception, we excluded implicit exceptions raised
+        # by executing RAISE_VARARGS and RERAISE, Cyberbrain can't handle exceptions
+        # caused by `raise 1` (TypeError). But since this should be super rare, it
+        # should be fine.
+
         cause = exc = None
         if instr.arg == 2:
             cause, exc = self._pop(2)
@@ -766,7 +902,8 @@ class Py37ValueStack(GeneralValueStack):
         self.why = Why.EXCEPTION
         self._fast_block_end()
 
-    def _WITH_CLEANUP_START_handler(self):
+    # noinspection DuplicatedCode
+    def _WITH_CLEANUP_START_handler(self, exc_info):
         exc = self.tos
         exit_func: any
         if not exc:  # Checks if tos is None, which in our stack, is []
@@ -776,13 +913,24 @@ class Py37ValueStack(GeneralValueStack):
                 exit_func = self.stack.pop(-2)  # why, ret_val, __exit__
             else:
                 exit_func = self.stack.pop(-1)  # why, __exit__
+        elif utils.is_exception_class(exc):
+            w, v, u = self._pop(3)
+            tp2, exc2, tb2 = self._pop(3)
+            exit_func = self._pop()
+            self._push(tp2, exc2, tb2)
+            self._push(None)
+            self._push(w, v, u)
+            block = self.block_stack.tos
+            assert block.b_type == BlockType.EXCEPT_HANDLER
+            block.b_level -= 1
         else:
-            assert False, "Unhandled exception is not supported by Cyberbrain"
+            assert False, f"Unrecognized type: {exc}"
 
-        self._push(_placeholder)  # exc, set to None
-        self._push(exit_func)
+        if self._instruction_successfully_executed(exc_info, "WITH_CLEANUP_START"):
+            self._push(exc)
+            self._push(exit_func)
 
-    def _END_FINALLY_handler(self):
+    def _END_FINALLY_handler(self, instr):
         status = self._pop()
         if isinstance(status, Why):
             self.why = status
@@ -791,9 +939,10 @@ class Py37ValueStack(GeneralValueStack):
                 self.return_value = self._pop()
             if self.why is Why.SILENCED:
                 block = self._pop_block()
-                assert block.type is BlockType.EXCEPT_HANDLER
+                assert block.b_type is BlockType.EXCEPT_HANDLER
                 self._unwind_except_handler(block)
                 self.why = Why.NOT
+                return
             self._fast_block_end()
         elif utils.is_exception_class(status):
             exc_type = status
@@ -865,6 +1014,35 @@ class Py38ValueStack(Py37ValueStack):
     This allows us to only override the methods that have changed in 3.8
     """
 
+    def _store_exception(self, exc_info: ExceptionInfo):
+        """When an exception is raised implicitly (aka not by calling `raise`), use
+        This method to propagate it as self.last_exception.
+
+        TODO: Every instruction handler that may raise exceptions should call this
+            method.
+        """
+        self.last_exception = exc_info
+        self._exception_unwind()
+
+    def _WITH_CLEANUP_FINISH_handler(self):
+        # For __exit__, returning a true value from this method will cause the with
+        # statement to suppress the exception and continue execution with the statement
+        # immediately following the with statement. Otherwise the exception continues
+        # propagating after this method has finished executing.
+        #
+        # res represents the return value, but there's no way CB can know its value.
+        # So we just assume res is true whenever there is an exception, because CB does
+        # not support unhandled exception, so it's safe to assume that if there's an
+        # exception raised in `with`, it is properly handled. e.g.
+        #         with pytest.raises(TypeError)
+        res = self._pop()
+        exc = self._pop()
+        if res and utils.is_exception(exc):
+            block = self.block_stack.pop()
+            assert block.b_type == BlockType.EXCEPT_HANDLER
+            self._unwind_except_handler(block)
+            self._push(NULL)
+
     def _RETURN_VALUE_handler(self):
         self.return_value = self._pop()
         # TODO: add exit_returning
@@ -885,7 +1063,7 @@ class Py38ValueStack(Py37ValueStack):
         # out of the switch clause, then jumps to label "error", which is above
         # _exception_unwind. So _exception_unwind will be executed anyway.
         self._do_raise(exc, cause)
-        self._exception_unwind(instr)
+        self._exception_unwind()
 
     def _POP_EXCEPT_handler(self):
         block = self._pop_block()
@@ -902,7 +1080,7 @@ class Py38ValueStack(Py37ValueStack):
             res = self._pop()
 
         if self.tos is NULL or isinstance(self.tos, int):
-            exc = self._pop()
+            _ = self._pop()
         else:
             _, _, _ = self._pop(3)
             block = self._pop_block()
@@ -932,26 +1110,30 @@ class Py38ValueStack(Py37ValueStack):
             self.last_exception = ExceptionInfo(
                 type=exc_type, value=value, traceback=tb
             )
-            self._exception_unwind(instr)
+            self._exception_unwind()
         else:
             raise ValueStackException(f"TOS has wrong value: {self.tos}")
 
-    def _WITH_CLEANUP_START_handler(self):
+    def _WITH_CLEANUP_START_handler(self, exc_info):
+        exc = self.tos
         if self.tos == NULL:  # Pushed by BEGIN_FINALLY
             exit_func = self.stack.pop(-2)
-            self._push([])  # Pushes None to the stack
-            self._push(exit_func)
         else:
-            # Note that in real CPython code, this clause is used to process unhandled
-            # exceptions thrown inside `with`. However, since our stack is just a
-            # simulation, we won't push exceptions onto the stack even if an unhandled
-            # happened, which, makes it meaningless to write any code here.
-            # Essentially, users shouldn't use Cyberbrain if there will ever be
-            # unhandled exceptions.
-            pass
+            w, v, u = self._pop(3)
+            tp2, exc2, tb2 = self._pop(3)
+            exit_func = self._pop()
+            self._push(tp2, exc2, tb2)
+            self._push(None)
+            self._push(w, v, u)
+            block = self.block_stack.tos
+            assert block.b_type == BlockType.EXCEPT_HANDLER
+            block.b_level -= 1
 
-    def _exception_unwind(self, instr):
-        print("inside _exception_unwind")
+        if self._instruction_successfully_executed(exc_info, "WITH_CLEANUP_START"):
+            self._push(exc)
+            self._push(exit_func)
+
+    def _exception_unwind(self):
         while self.block_stack.is_not_empty():
             block = self.block_stack.pop()
 
@@ -977,11 +1159,11 @@ class Py39ValueStack(Py38ValueStack):
     def _JUMP_IF_NOT_EXC_MATCH_handler(self):
         self._pop(2)
 
-    def _CONTAINS_OP_handler(self):
-        self._BINARY_operation_handler()
+    def _CONTAINS_OP_handler(self, exc_info):
+        self._BINARY_operation_handler(exc_info)
 
-    def _IS_OP_handler(self):
-        self._BINARY_operation_handler()
+    def _IS_OP_handler(self, exc_info):
+        self._BINARY_operation_handler(exc_info)
 
     def _LOAD_ASSERTION_ERROR_handler(self):
         self._push(AssertionError())
@@ -989,30 +1171,38 @@ class Py39ValueStack(Py38ValueStack):
     def _LIST_TO_TUPLE_handler(self, instr):
         pass
 
-    def _LIST_EXTEND_handler(self):
+    def _LIST_EXTEND_handler(self, exc_info):
         # list.extend(TOS1[-i], TOS), which essentially merges tos and tos1.
-        self._pop_n_push_one(2)
+        items = self._pop(2)
+        if self._instruction_successfully_executed(exc_info, "LIST_EXTEND"):
+            self._push(utils.flatten(items))
 
-    def _SET_UPDATE_handler(self):
+    def _SET_UPDATE_handler(self, exc_info):
         # list.extend(TOS1[-i], TOS), which essentially merges tos and tos1.
-        self._pop_n_push_one(2)
+        items = self._pop(2)
+        if self._instruction_successfully_executed(exc_info, "SET_UPDATE"):
+            self._push(utils.flatten(items))
 
-    def _DICT_UPDATE_handler(self):
+    def _DICT_UPDATE_handler(self, exc_info):
         # dict.extend(TOS1[-i], TOS), which essentially merges tos and tos1.
-        self._pop_n_push_one(2)
+        items = self._pop(2)
+        if self._instruction_successfully_executed(exc_info, "DICT_UPDATE"):
+            self._push(utils.flatten(items))
 
-    def _DICT_MERGE_handler(self):
-        self._pop_n_push_one(2)
+    def _DICT_MERGE_handler(self, exc_info):
+        items = self._pop(2)
+        if self._instruction_successfully_executed(exc_info, "DICT_MERGE"):
+            self._push(utils.flatten(items))
 
-    def _RERAISE_handler(self, instr):
+    def _RERAISE_handler(self):
         exc_type = self._pop()
         value = self._pop()
         tb = self._pop()
         assert utils.is_exception_class(exc_type)
         self.last_exception = ExceptionInfo(type=exc_type, value=value, traceback=tb)
-        self._exception_unwind(instr)
+        self._exception_unwind()
 
-    def _WITH_EXCEPT_START_handler(self, instr):
+    def _WITH_EXCEPT_START_handler(self):
         exit_func = self.stack[-7]
         self._push(exit_func)
 
